// Copyright (c) 2013 by István Váradi

// This file is part of XPLRA, a remote-access plugin for X-Plane

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice, this
//    list of conditions and the following disclaimer.
// 2. Redistributions in binary form must reproduce the above copyright notice,
//    this list of conditions and the following disclaimer in the documentation
//    and/or other materials provided with the distribution.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
// ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

// The views and conclusions contained in the software and documentation are those
// of the authors and should not be interpreted as representing official policies,
// either expressed or implied, of the FreeBSD Project.

//------------------------------------------------------------------------------

#include "xplra.h"

#include "XPlane.h"
#include "Exception.h"

#include <vector>
#include <memory>

#include <cstdio>
#include <cassert>

//------------------------------------------------------------------------------

using hu::varadiistvan::xplra::XPlane;
using hu::varadiistvan::xplra::Exception;
using hu::varadiistvan::xplra::IOException;
using hu::varadiistvan::xplra::ProtocolException;
using hu::varadiistvan::xplra::NotConnectedException;
using hu::varadiistvan::xplra::TypeMismatchException;
using hu::varadiistvan::xplra::InvalidIDException;

using std::exception;
using std::string;
using std::vector;
using std::auto_ptr;

//------------------------------------------------------------------------------

namespace {

//------------------------------------------------------------------------------

/**
 * Information about a connection.
 */
class Connection : public XPlane
{
private:
    /**
     * The last error code.
     */
    int lastErrorCode;

    /**
     * The last error subcode.
     */
    unsigned long lastErrorSubCode;

    /**
     * The string of the last error
     */
    std::string lastErrorString;

public:
    /**
     * Construct the connection
     */
    Connection() throw();

    /**
     * Handle the currently pending exception.
     */
    void handleException() throw();

    /**
     * Get the last error code.
     */
    int getLastError(unsigned long* lastErrorSubCode) const throw();

    /**
     * Get the string representation of the last error code.
     */
    const char* getLastErrorString() const throw();

    /**
     * Clear the last error code.
     */
    void clearLastError() throw();
};

//------------------------------------------------------------------------------

inline Connection::Connection() throw() :
    lastErrorCode(0),
    lastErrorSubCode(0)
{
}

//------------------------------------------------------------------------------

void Connection::handleException() throw()
{
    try {
        throw;
    } catch(const IOException& e) {
        lastErrorCode = ERROR_IO;
        lastErrorSubCode = e.getErrorCode();
        lastErrorString = e.what();
    } catch(const ProtocolException& e) {
        lastErrorCode = ERROR_PROTOCOL;
        lastErrorSubCode = static_cast<unsigned long>(e.getErrorCode());
        lastErrorString = e.what();
    } catch(const NotConnectedException& e) {
        lastErrorCode = ERROR_NOT_CONNECTED;
        lastErrorSubCode = 0;
        lastErrorString = e.what();
    } catch(const TypeMismatchException& e) {
        lastErrorCode = ERROR_TYPE_MISMATCH;
        lastErrorSubCode = 0;
        lastErrorString = e.what();
    } catch(const InvalidIDException& e) {
        lastErrorCode = ERROR_INVALID_ID;
        lastErrorSubCode = 0;
        lastErrorString = e.what();
    } catch(const exception& e) {
        lastErrorCode = ERROR_OTHER;
        lastErrorSubCode = 0;
        lastErrorString = e.what();
    } catch(...) {
        lastErrorCode = ERROR_OTHER;
        lastErrorSubCode = 0;
        lastErrorString = "<Exception of an unknown type>";
    }
}

//------------------------------------------------------------------------------

inline int Connection::getLastError(unsigned long* lastErrorSubCode)
    const throw()
{
    if (lastErrorSubCode!=0) *lastErrorSubCode = this->lastErrorSubCode;
    return lastErrorCode;
}

//------------------------------------------------------------------------------

inline const char* Connection::getLastErrorString() const throw()
{
    return (lastErrorCode==ERROR_NONE) ? 0 : lastErrorString.c_str();
}

//------------------------------------------------------------------------------

inline void Connection::clearLastError() throw()
{
    lastErrorCode = ERROR_NONE;
    lastErrorSubCode = 0;
    lastErrorString.clear();
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

/**
 * Information about a connection slot in the vector of connections.
 */
struct Slot
{
private:
    /**
     * The slots.
     */
    static std::vector<Slot> slots;

    /**
     * The index of the first free slot.
     */
    static int firstFree;

public:
    /**
     * Add a new connection.
     */
    static int addConnection(Connection* connection) throw();

    /**
     * Get the connection with the given ID.
     */
    static Connection* getConnection(int connectionID) throw();

    /**
     * Clear the connection with the given ID.
     */
    static void clearConnection(int connectionID) throw();

    // Indicate if the slot contains a valid connection or not
    bool valid;

    union {
        // The connection, if it is a valid connection
        Connection* connection;

        // The index of the next free slot
        int nextFreeIndex;
    };

    /**
     * Construct a slot with the given connection
     */
    Slot(Connection* connection) throw();

    /**
     * Set the connection and return the former next free index. It
     * should be called for a slot with no connection.
     */
    int setConnection(Connection* connection) throw();

    /**
     * Get the connection if this slot contains a connection.
     */
    Connection* getConnection() const throw();

    /**
     * Clear the connection and set the given next free index.
     */
    void clear(int nextFreeIndex) throw();
};

//------------------------------------------------------------------------------

vector<Slot> Slot::slots;

//------------------------------------------------------------------------------

int Slot::firstFree = -1;

//------------------------------------------------------------------------------

inline Slot::Slot(Connection* connection) throw() :
    valid(true)
{
    this->connection = connection;
}

//------------------------------------------------------------------------------

inline int Slot::setConnection(Connection* connection) throw()
{
    assert(!valid);
    int nextFreeIndex = this->nextFreeIndex;
    this->connection = connection;
    valid = true;
    return nextFreeIndex;
}

//------------------------------------------------------------------------------

inline Connection* Slot::getConnection() const throw()
{
    return valid ? connection : 0;
}

//------------------------------------------------------------------------------

inline void Slot::clear(int nextFreeIndex) throw()
{
    assert(valid);
    valid = false;
    this->nextFreeIndex = nextFreeIndex;
}

//------------------------------------------------------------------------------

int Slot::addConnection(Connection* connection) throw()
{
    int id = firstFree;
    if (id<0) {
        id = slots.size();
        slots.push_back(Slot(connection));
    } else {
        Slot& slot = slots[id];
        firstFree = slot.setConnection(connection);
    }

    return id;
}

//------------------------------------------------------------------------------

Connection* Slot::getConnection(int connectionID) throw()
{
    size_t index = static_cast<size_t>(connectionID);
    return (index<slots.size()) ? slots[index].getConnection() : 0;
}

//------------------------------------------------------------------------------

void Slot::clearConnection(int connectionID) throw()
{
    size_t index = static_cast<size_t>(connectionID);
    if (index<slots.size()) {
        slots[index].clear(firstFree);
        firstFree = index;
    }
}

//------------------------------------------------------------------------------

} /* anonymous namespace */

//------------------------------------------------------------------------------

extern "C" int xplra_get_last_error(int connectionID, unsigned long* subCode)
{
    Connection* connection = Slot::getConnection(connectionID);
    return (connection==0) ? -1 : connection->getLastError(subCode);
}

//------------------------------------------------------------------------------

extern "C" const char* xplra_get_last_error_string(int connectionID)
{
    Connection* connection = Slot::getConnection(connectionID);
    return (connection==0) ? 0 : connection->getLastErrorString();
}

//------------------------------------------------------------------------------

extern "C" void xplra_clear_last_error(int connectionID)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection!=0) connection->clearLastError();
}

//------------------------------------------------------------------------------

extern "C" int xplra_connect()
{
    try {
        auto_ptr<Connection> connection(new Connection());
        connection->connect();
        return Slot::addConnection(connection.release());
    } catch(...) {
        return -1;
    }
}

//------------------------------------------------------------------------------

extern "C" int xplra_get_int(int* value, int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        *value = connection->getInt(name);
        return 0;
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

//------------------------------------------------------------------------------

extern "C" int xplra_get_float(float* value, int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        *value = connection->getFloat(name);
        return 0;
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

//------------------------------------------------------------------------------

extern "C" int xplra_get_double(double* value,
                                int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        *value = connection->getDouble(name);
        return 0;
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

//------------------------------------------------------------------------------

extern "C" ssize_t xplra_get_float_array(float* dest, size_t length, size_t offset,
                                         int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        return connection->getFloatArray(name, dest, length, offset);
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

/*----------------------------------------------------------------------------*/

extern "C" float* xplra_get_float_array_new(size_t* length, size_t offset,
                                            int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return 0;
    try {
        return connection->getFloatArray(name, *length, offset);
    } catch (...) {
        connection->handleException();
        return 0;
    }
}

//------------------------------------------------------------------------------

extern "C" ssize_t xplra_get_int_array(int32_t* dest, size_t length, size_t offset,
                                       int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        return connection->getIntArray(name, dest, length, offset);
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

/*----------------------------------------------------------------------------*/

extern "C" int32_t* xplra_get_int_array_new(size_t* length, size_t offset,
                                            int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return 0;
    try {
        return connection->getIntArray(name, *length, offset);
    } catch (...) {
        connection->handleException();
        return 0;
    }
}

//------------------------------------------------------------------------------

extern "C" ssize_t xplra_get_byte_array(void* dest, size_t length, size_t offset,
                                        int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;
    try {
        return connection->getByteArray(name, reinterpret_cast<uint8_t*>(dest),
                                        length, offset);
    } catch (...) {
        connection->handleException();
        return -1;
    }
}

/*----------------------------------------------------------------------------*/

extern "C" uint8_t* xplra_get_byte_array_new(size_t* length, size_t offset,
                                             int connectionID, const char* name)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return 0;
    try {
        return connection->getByteArray(name, *length, offset);
    } catch (...) {
        connection->handleException();
        return 0;
    }
}

//------------------------------------------------------------------------------

extern "C" int xplra_disconnect(int connectionID)
{
    Connection* connection = Slot::getConnection(connectionID);
    if (connection==0) return -1;

    Slot::clearConnection(connectionID);
    connection->disconnect();
    delete connection;

    return 0;
}

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

// Local Variables:
// mode: C++
// c-basic-offset: 4
// indent-tabs-mode: nil
// End:
